r"""
Contains the objects for computing the Riemann Sum of functions of several variables over an
arbitrary number of dimensions over a given interval.

.. py:data:: LEFT

    Specifies that the Left Riemann Sum method should be used over the corresopnding dimension.

    .. math::

        x_{i}^{*} = a+i\Delta x, \Delta x=\frac{b-a}{n}, i \in \{0,1,\dots,n-1\}

    :type: Method

.. py:data:: MIDDLE

    Specifices that the Middle Riemann Sum method should be used over the corresponding dimension.

    .. math::

        x_{i}^{*} = a+\frac{2i+1}{2}\Delta x, \Delta x=\frac{b-a}{n}, i \in \{0,1,\dots,n-1\}

    :type: Method

.. py:data:: RIGHT

    Specifies that the Right Riemann Sum method should be used over the corresopnding dimension.

    .. math::

        x_{i}^{*} = a+(i+1)\Delta x, \Delta x = \frac{b-a}{n}, i \in \{0,\dots,n-1\}

    :type: Method
"""

from dataclasses import dataclass
from decimal import Decimal
import inspect
import itertools
from numbers import Number
import typing


@dataclass
class Interval:
    """
    Contains the bounds of an interval.

    .. py:attribute:: lower

        The lower bound of the interval.

        :type: :py:class:`decimal.Decimal`

    .. py:attribute:: upper

        The upper bound of the interval.

        :type: :py:class:`decimal.Decimal`
    """
    lower: Decimal
    upper: Decimal


@dataclass
class Method:
    """
    Abstract class for defining custom Riemann Sum methods. There is built-in support for the three
    most common Riemann Sum methods:

    - Left Riemann Sum method: :py:data:`riemann.LEFT`
    - Middle Riemann Sum method: :py:data:`riemann.MIDDLE`
    - Right Riemann Sum method: :py:data:`riemann.RIGHT`

    .. py:attribute:: name

        The name of the Riemann Sum method. This attribute is arbitrary and is solely used when
        representing a :py:class:`riemann.Method` object as a string (e.g., for debugging
        purposes).

        :type: str

    .. py:attribute:: func

        A callable object that takes three parameters:

        1. An :py:class:`riemann.Interval` object
        2. An ``int`` object
        3. A :py:class:`decimal.Decimal` object.

        Yields the value of the independent variable at the :math:`i`-th partition in the closed
        interval :math:`[a, b]`. 

        :type: typing.Callable
    """
    name: str
    func: typing.Callable[[Interval, int, Decimal], Decimal]

    def __repr__(self) -> str:
        """
        """
        return f"Method(name='{self.name}')"

    def partitions(
        self, interval: Interval, n: int, delta: Decimal
    ) -> typing.Generator[Decimal, None, None]:
        """
        Computes the values of the independent variable at each of the :math:`n` partitions in the
        closed interval :math:`[a, b]`.

        :param interval: The (closed) interval of the summation
        :param n: The number of partitions into which the interval of the summation is divided
        :param delta: The length of the each of the :math:`n` partition
        """
        return (self.func(interval, i, delta) for i in range(n))


LEFT = Method("left", lambda x, i, d: x.lower + i * d)
MIDDLE = Method("middle", lambda x, i, d: x.lower + Decimal(2 * i + 1) / 2 * d)
RIGHT = Method("right", lambda x, i, d: x.lower + (i + 1) * d)


@dataclass
class Dimension:
    """
    Contains the parameters of the summation over the dimension of interest.

    .. py:attribute:: a

        The lower bound of the interval of the summation.

        :type: numbers.Number

    .. py:attribute:: b

        The upper bound of the interval of the summation.

        :type: number.Number

    .. py:attribute:: n

        The number of partitions into which the interval of the summation :math:`[a, b]` is
        divided.

        :type: int

    .. py:attribute:: method

        The Riemann Sum method to use.

        :type: Method
    """
    a: Number
    b: Number
    n: int
    method: Method


def rsum(func: typing.Callable[..., Number], *args: Dimension):
    r"""
    Computes the Riemann Sum of functions of several variables over an arbitrary number of
    dimensions over a given interval.

    Parameter ``func`` can be written as :math:`f: {\mathbb{R}}^{n} \rightarrow \mathbb{R}`. The
    number of items in ``args`` must equal :math:`n`, the number of parameters of ``func`` and the
    number of dimensions of :math:`f`.

    :param func: A function of several real variables
    :param args: The parameters of the summation over each of the dimensions
    :return: The value of the Riemann Sum
    :raise ValueError: The number of dimensions does not equal the number of parameters of ``func``
    """
    if len(args) != len(inspect.signature(func).parameters):
        raise ValueError(
            "The number of values in 'args' does not equal the number of parameters of 'func'"
        )

    # Contains generators that yield the values to pass to the ``func``.
    # Each element represents one of the :math:`n` dimensions.
    values = []

    # :math:`\Delta V_{i}`
    delta = Decimal(1)

    # Iterate through the :math:`n` dimensions
    for dim in args:
        # Create :py:class:`Interval` object from bounds
        interval = Interval(
            Decimal(str(dim.a) if isinstance(dim.a, float) else dim.a),
            Decimal(str(dim.b) if isinstance(dim.b, float) else dim.b)
        )

        # Compute :math:`\Delta x` for the :math:`n`-th dimension.
        dvar = (interval.upper - interval.lower) / dim.n
        delta *= dvar

        values.append(dim.method.partitions(interval, dim.n, dvar))

    # Compute the :math:`n`-th dimensional Riemann Sum.
    return (delta * sum(func(*v) for v in itertools.product(*values))).normalize()



"""
Tests for :py:mod:`riemann.riemann`.
"""

from decimal import Decimal
import typing

import pytest

from riemann import riemann

# Type aliases
D = Decimal
Dim = riemann.Dimension

# Global variable aliases
L, M, U = riemann.LOWER, riemann.MIDDLE, riemann.UPPER


@pytest.mark.parametrize(
    "func, dimensions, x", [
        (lambda x: 0, [Dim(0, 1, 1, L)], 0),
        (lambda x: 0, [Dim(0, 1, 1, M)], 0),
        (lambda x: 0, [Dim(0, 1, 1, U)], 0),

        (lambda x: 1, [Dim(0, 1, 1, L)], 1),
        (lambda x: 1, [Dim(0, 1, 1, M)], 1),
        (lambda x: 1, [Dim(0, 1, 1, U)], 1),

        (lambda x: x, [Dim(0, 1, 1, L)], 0),
        (lambda x: x, [Dim(0, 1, 1, M)], 1 / D(2)),
        (lambda x: x, [Dim(0, 1, 1, U)], 1),

        (lambda x: x ** 2, [Dim(0, 1, 1, L)], 0),
        (lambda x: x ** 2, [Dim(0, 1, 1, M)], 1 / D(4)),
        (lambda x: x ** 2, [Dim(0, 1, 1, U)], 1),
        (lambda x: x ** 2, [Dim(0, 1, 2, L)], 1 / D(8)),
        (lambda x: x ** 2, [Dim(0, 1, 2, M)], D(5) / D(16)),
        (lambda x: x ** 2, [Dim(0, 1, 2, U)], D(5) / D(8)),
        (lambda x: x ** 2, [Dim(0, 2, 2, L)], 1),
        (lambda x: x ** 2, [Dim(0, 2, 2, M)], D(5) / D(2)),
        (lambda x: x ** 2, [Dim(0, 2, 2, U)], 5),
        (lambda x: x ** 2, [Dim(0, 2, 1, L)], 0),
        (lambda x: x ** 2, [Dim(0, 2, 1, M)], 2),
        (lambda x: x ** 2, [Dim(0, 2, 1, U)], 8),
        (lambda x: x ** 2, [Dim(1, 6, 10, L)], D(505) / D(8)),
        (lambda x: x ** 2, [Dim(1, 6, 10, M)], D(1145) / D(16)),
        (lambda x: x ** 2, [Dim(1, 6, 10, U)], D(645) / D(8)),
        (lambda x: x ** 2, [Dim(-10, 10, 50, L)], D(3336) / D(5)),
        (lambda x: x ** 2, [Dim(-10, 10, 50, M)], D(3332) / D(5)),
        (lambda x: x ** 2, [Dim(-10, 10, 50, U)], D(3336) / D(5)),

        (lambda x: x ** 3, [Dim(0, 1, 1, L)], 0),
        (lambda x: x ** 3, [Dim(0, 1, 1, M)], 1 / D(8)),
        (lambda x: x ** 3, [Dim(0, 1, 1, U)], 1),

        (lambda x: x ** 4 + x ** 2 - 4 * x + 3, [Dim(-5, 5, 25, L)], D(865578) / D(625)),
        (lambda x: x ** 4 + x ** 2 - 4 * x + 3, [Dim(-5, 5, 25, M)], D(847838) / D(625)),
        (lambda x: x ** 4 + x ** 2 - 4 * x + 3, [Dim(-5, 5, 25, U)], D(855578) / D(625)),
    ]
)
def test_summation_1d(func: typing.Callable, dimensions: typing.Iterable[Dim], x: D):
    """

    :param func:
    :param dimensions:
    :param x:
    :return:
    """
    assert riemann.rsum(func, *dimensions) == x


@pytest.mark.parametrize(
    "func, dimensions, x", []
)
def test_summation_2d(func: typing.Callable, dimensions: typing.Iterable[Dim], x: D):
    """

    :param func:
    :param dimensions:
    :param x:
    :return:
    """
    assert riemann.rsum(func, *dimensions) == x



r"""
Computes the Riemann Sum of functions of several variables over an arbitrary number of dimensions
over a given interval.

From Wikipedia:

    Let :math:`f: [a,b] \rightarrow \mathbb{R}` be a function defined on a closed interval
    :math:`[a,b]` of the real numbers, :math:`\mathbb(R)`, and

    .. math::

        P = \{[x_{0},x_{1}], [x_{1},x_{2}], \dots , [x_{n-1},x_{n}]\},

    be a partition of :math:`I`, where

    .. math::

        a = x_{0} < x_{1} < x_{2} < \dots < x_{n} = b.

    A Riemann sum of :math:`S` of :math:`f` over :math:`I` with partition :math:`P` is defined as

    .. math::

        S = \sum_{i=1}^{n} f(x_{i}^{*}) \Delta x_{i}

    where :math:`\Delta x_{i} = x_{i} - x_{i-1}` and :math:`x_{i}^{*} \in [x_{i-1},x_{i}]`.

(Source: `[1] <https://en.wikipedia.org/wiki/Riemann_sum#Definition>`_)

From Wikipedia:

    Higher dimensional Riemann sums follow a similar pattern as from one to two to three dimensions.
    For an arbitrary dimension, :math:`n`, a Riemann sum can be written as

    .. math::

        S = \sum_{i=1}^{n} f(P_{i}^{*}) \Delta V_{i}

    where :math:`P_{i}^{*} \in V_{i}`, that is, it's a point in the :math:`n`-dimensional cell
    :math:`V_{i}` with :math:`n`-dimensional volume :math:`\Delta V_{i}`.

(Source: `[2] <https://en.wikipedia.org/wiki/Riemann_sum#Arbitrary_number_of_dimensions>`_)
"""

from .riemann import LEFT, MIDDLE, RIGHT
from .riemann import Dimension
from .riemann import Interval
from .riemann import Method
from .riemann import rsum
